"""
Utils
"""

from functools import partial
from typing import Mapping, KT, VT, Iterable, Callable, List
from collections import deque


def identity(x):
    return x


def split_keys(d):
    """Returns a dictionary where keys that had spaces were split into multiple keys

    Meant to be a convenience function for the user to use when they want to define a
    mapping where several keys map to the same value.

    >>> split_keys({'apple': 1, 'banana carrot': 2})
    {'apple': 1, 'banana': 2, 'carrot': 2}
    """
    return {split_k: v for k, v in d.items() for split_k in k.split()}


from typing import Dict, KT, Literal


def intersection_graph(
    sets: Dict[KT, set], edge_labels: Literal['elements', 'size', False] = False
):
    """
    A graph of all intersections between sets.
    (See https://en.wikipedia.org/wiki/Intersection_graph.)

    In graph theory, an adjacency list is a collection of sets used to represent a
    finite graph.
    Here, the vertices are the values of sets,
    and there is an edge between two vertices if the sets intersect.
    The weight of the edge is the size of the intersection.

    :param sets: A mapping of keys to sets of elements. These sets of elements will
        be the vertices of the graph.
    :param edge_labels: If 'elements', the edge labels are the elements of the intersection.
        If 'size', the edge labels are the size of the intersection.
        If False, there are no edge labels.
    :return: A graph, represented by an "adjacency list"
        (see https://en.wikipedia.org/wiki/Adjacency_list)
        (a dict whose keys are the keys of the input `sets` dict, and whose values
        tell us what sets of `sets` intersect with it),
        optionally with some information about this intersection.

    >>> sets = {
    ...     'A': {'b', 'c'},
    ...     'B': {'a', 'b', 'd', 'e', 'f'},
    ...     'C': {'f', 'g'},
    ...     'D': {'d', 'e', 'h', 'i'},
    ...     'E': {'i', 'j'}
    ... }
    >>> assert intersection_graph(sets) == {
    ...     'A': {'B'}, 'B': {'A', 'C', 'D'}, 'C': {'B'}, 'D': {'B', 'E'}, 'E': {'D'}
    ... }
    >>> assert intersection_graph(sets, edge_labels='elements') == {
    ...     'A': {'B': {'b'}},
    ...     'B': {'A': {'b'}, 'C': {'f'}, 'D': {'d', 'e'}},
    ...     'C': {'B': {'f'}},
    ...     'D': {'B': {'d', 'e'}, 'E': {'i'}},
    ...     'E': {'D': {'i'}}
    ... }
    >>> assert intersection_graph(sets, edge_labels='size') == {
    ...     'A': {'B': 1},
    ...     'B': {'A': 1, 'C': 1, 'D': 2},
    ...     'C': {'B': 1},
    ...     'D': {'B': 2, 'E': 1},
    ...     'E': {'D': 1}
    ... }

    """
    from collections import defaultdict
    from itertools import combinations

    graph = defaultdict(dict)
    for key1, key2 in combinations(sets.keys(), 2):
        intersection = sets[key1] & sets[key2]
        if intersection:
            graph[key1][key2] = intersection
            graph[key2][key1] = intersection
    if isinstance(edge_labels, str):
        if edge_labels == 'elements':
            return dict(graph)
        elif edge_labels == 'size':
            return map_values(map_values.len, graph)
    elif edge_labels is False:
        return map_values.set(graph)

    raise ValueError(f'Invalid value for edge_labels: {edge_labels}')


def map_values(func: Callable, d: dict):
    """Apply a function to all values of a dictionary.

    >>> map_values(lambda x: x ** 2, {1: 2, 3: 4})
    {1: 4, 3: 16}
    """
    return {k: func(v) for k, v in d.items()}


map_values.len = partial(map_values, len)
map_values.set = partial(map_values, set)
map_values.list = partial(map_values, list)


def invert_labeled_collection(
    d: Mapping[KT, Iterable[VT]], values_container: Callable = list
) -> Mapping[VT, Iterable[KT]]:
    """Invert a mapping whose values are iterables of objects,
    getting a mapping from objects to iterables of keys.

    >>> original_dict = {
    ...     "X": ['a', 'b'],
    ...     "Y": ['a'],
    ...     "Z": ['a', 'b', 'c']
    ... }
    >>> inverted_dict = invert_labeled_collection(original_dict)
    >>> inverted_dict
    {'a': ['X', 'Y', 'Z'], 'b': ['X', 'Z'], 'c': ['Z']}
    >>> invert_labeled_collection(inverted_dict)
    {'X': ['a', 'b'], 'Y': ['a'], 'Z': ['a', 'b', 'c']}

    The `values_container` argument can be used to cast the values of the inverted dict.

    >>> assert (
    ...     invert_labeled_collection(original_dict, values_container=set)
    ...     == {'a': {'X', 'Y', 'Z'}, 'b': {'X', 'Z'}, 'c': {'Z'}}
    ... )
    >>>
    >>> d = {'a': 'apple', 'b': 'banana'}
    >>> t = invert_labeled_collection(d, values_container=''.join)
    >>> t
    {'a': 'abbb', 'p': 'aa', 'l': 'a', 'e': 'a', 'b': 'b', 'n': 'bb'}
    >>> invert_labeled_collection(t, ''.join)
    {'a': 'apple', 'b': 'aaabnn'}

    """
    inverted_dict = {}
    for key, values in d.items():
        for value in values:
            inverted_dict.setdefault(value, []).append(key)
    if values_container is not list:
        inverted_dict = {k: values_container(v) for k, v in inverted_dict.items()}
    return inverted_dict


def breadth_first_traversal(graph: Mapping, start_node, *, yield_edges=False):
    """Yields nodes starting from the root node, expanding to neighbors recursively,
    using breadth-first search, without repeating nodes.

    :param graph: Adjacencies of the graph: A mapping from nodes to their neighbors.
    :param start_node: The node to start from (key of the graph adjacency mapping)
    :param yield_edges: If True, yield edges instead of nodes.
        The edges are yielded as tuples of (node, neighbor).

    >>> graph = {
    ...     'A': ['B'], 'B': ['A', 'C', 'D'], 'C': ['B'], 'D': ['B', 'E'], 'E': ['D']
    ... }
    >>> list(breadth_first_traversal(graph, 'B'))
    ['B', 'A', 'C', 'D', 'E']
    >>> list(breadth_first_traversal(graph, 'B', yield_edges=True))
    [('B', 'A'), ('B', 'C'), ('B', 'D'), ('D', 'E')]

    """

    visited = set()
    queue = deque([start_node])

    while queue:
        node = queue.popleft()
        if node not in visited:
            if yield_edges:
                for neighbor in graph[node]:
                    if neighbor not in visited:
                        yield (node, neighbor)
            else:
                yield node
            visited.add(node)
            queue.extend(
                neighbor for neighbor in graph[node] if neighbor not in visited
            )


# TODO: Integrate in tabled.base.dflt_ext_mapping
def auto_decode_bytes(
    b: bytes, *, try_first_bytes=(1e6, 1e7, 1e8), encoding: str = 'utf-8', verbose=False
) -> str:
    """
    Decode a byte sequence into a string, trying charset_normalizer gueses if fails.

    This function attempts to decode the given bytes using the default encoding (usually 'utf-8').
    If that fails due to a `UnicodeDecodeError`, it uses `charset_normalizer` to detect the encoding
    by analyzing increasingly larger samples of the byte sequence, as specified in `try_first_bytes`.
    If all attempts fail, it analyzes the entire byte sequence to detect the encoding.

    Parameters:
        b (bytes): The byte sequence to decode.
        try_first_bytes (tuple of floats): Byte lengths to use for encoding detection samples.
            Defaults to (1e6, 1e7, 1e8).

    Returns:
        str: The decoded string.

    Raises:
        UnicodeDecodeError: If the byte sequence cannot be decoded after all attempts.

    Examples:
        >>> # Example with UTF-8 encoded bytes
        >>> s = 'Hello, world! Привет мир! こんにちは世界！'
        >>> b_utf8 = s.encode('utf-8')
        >>> auto_decode_bytes(b_utf8) == s
        True

        >>> # Example with UTF-16 encoded bytes
        >>> s_utf16 = 'Hello, world! 你好，世界！'
        >>> b_utf16 = s_utf16.encode('utf-16')
        >>> auto_decode_bytes(b_utf16) == s_utf16
        True

        Now, this is auto_decoding, but it doesn't mean it's robust.
        We use `charset_normalizer` to detect the encoding of the bytes, and then
        try to decode it with that encoding.
        But sometimes you can decode something that is not the original string,
        so be careful!!
        It's annoying to have to specify the encoding all the time, but this
        explicitness, and the errors that come with it, can be vital.

        Here are a few examples. We'll

        >>> s_latin1 = 'Héllo, wörld! Ça va?'
        >>> b_latin1 = s_latin1.encode('latin-1')  # latin-1 is ISO-8859-1
        >>> decoded_s = auto_decode_bytes(b_latin1, verbose=True)
        Trying encoding: 'utf-8'
        Trying encoding: 'big5'
        >>> decoded_s
        'H幨lo, w顤ld! ド va?'
        >>> decoded_s == s_latin1
        False

        >>> s_cp1252 = 'Special characters: € £ ¥ © ®'
        >>> b_cp1252 = s_cp1252.encode('cp1252')  # i.e. 'Windows-1252'
        >>> decoded_s = auto_decode_bytes(b_cp1252, verbose=True)
        Trying encoding: 'utf-8'
        Trying encoding: 'cp1125'
        >>> # See that charset_normalizer
        >>> decoded_s
        'Special characters: А г е й о'

    """

    import charset_normalizer  # pip install charset-normalizer

    if verbose:
        clog = lambda encoding: print(f"Trying encoding: '{encoding}'")
    else:
        clog = lambda encoding: None

    try:  # Try to decode using Python's default encoding (usually 'utf-8')
        clog(encoding)
        return b.decode(encoding)  # Use bytes.decode default encoding
    except UnicodeDecodeError:
        pass  # Proceed to detection steps

    import charset_normalizer  # pip install charset-normalizer

    try_first_bytes = list(map(int, try_first_bytes))

    for n_bytes in try_first_bytes:
        sample = b[:n_bytes]
        results = charset_normalizer.from_bytes(sample)
        if results:
            best_match = results.best()
            encoding = best_match.encoding
            try:
                clog(encoding)
                return b.decode(encoding=encoding)
            except UnicodeDecodeError:
                continue  # Try next sample size
        else:
            continue  # Detection failed, try next sample size

    # As a last resort, detect encoding using the entire byte sequence
    results = charset_normalizer.from_bytes(b)
    if results:
        encoding = results.best().encoding
        try:
            clog(encoding)
            return b.decode(encoding=encoding)
        except UnicodeDecodeError:
            pass  # Will raise error below

    raise UnicodeDecodeError('Unable to decode bytes with detected encodings.')


# -------------------------------------------------------------------------------------
# Expand and collapse rows and columns

import pandas as pd
from typing import List, Callable, Iterable, Union
import pandas as pd


def collapse_rows(
    df: pd.DataFrame, by: List[KT], *, container: Callable[[Iterable], Iterable] = list,
) -> pd.DataFrame:
    """
    Do a groupby to collapse (the rows of) a dataframe, gathering the other
    column's values (the ones that are not keys of the groupby) into lists.

    :param df: the dataframe to collapse
    :param by: the columns to group by (the keys of the groupby)
    :param container: the container to use to gather the other columns values

    >>> df = pd.DataFrame({
    ...     'a': [1, 1, 2, 2],
    ...     'b': [3, 4, 5, 6],
    ...     'c': [7, 8, 9, 10]
    ... })
    >>> df  # doctest: +NORMALIZE_WHITESPACE
       a  b   c
    0  1  3   7
    1  1  4   8
    2  2  5   9
    3  2  6  10
    >>> collapse_rows(df, ['a'])  # doctest: +NORMALIZE_WHITESPACE
       a       b        c
    0  1  [3, 4]   [7, 8]
    1  2  [5, 6]  [9, 10]

    """
    return df.groupby(by, as_index=False).agg(container)


def expand_rows(
    df: pd.DataFrame, grouped_columns: Union[str, List[KT]]
) -> pd.DataFrame:
    """
    Expands a DataFrame where specific columns were collapsed into containers back to its original form.
    Each column in `grouped_columns` should contain lists of the same length within each row.

    :param df: The DataFrame to expand.
    :param grouped_columns: The list of columns to expand
    :return: The expanded DataFrame.

    >>> df_collapsed = pd.DataFrame({
    ...     'a': [1, 2],
    ...     'b': [[3, 4], [5, 6, 66]],
    ...     'c': [[7, 8], [9, 10, 11]]
    ... })
    >>> expand_rows(df_collapsed, ['b', 'c'])  # doctest: +NORMALIZE_WHITESPACE
        a  b   c
    0  1  3   7
    1  1  4   8
    2  2  5   9
    3  2  6  10
    4  2  66  11
    """
    if isinstance(grouped_columns, str):
        grouped_columns = [grouped_columns]

    # Create a new list to hold the expanded rows
    expanded_rows = []

    # Iterate over each row in the DataFrame
    for _, row in df.iterrows():
        # Extract the lengths of lists in the grouped columns, assume all are the same length
        list_lengths = [len(row[col]) for col in grouped_columns]

        # Check if all list lengths are the same for safety
        if len(set(list_lengths)) != 1:
            raise ValueError(
                'All lists in a single row must have the same length to expand correctly.'
            )

        # Create individual rows for each index of the lists in grouped_columns
        for i in range(list_lengths[0]):
            # Start with a dict of non-grouped data
            new_row = {
                col: row[col] for col in df.columns if col not in grouped_columns
            }
            # Add the i-th element of each grouped column's list to the new row
            new_row.update({col: row[col][i] for col in grouped_columns})
            expanded_rows.append(new_row)

    # Create a new DataFrame from the expanded rows
    return pd.DataFrame(expanded_rows)


def collapse_columns(
    df: pd.DataFrame, groupings: Union[Dict[str, List[str]], Union[str, List[KT]]],
) -> pd.DataFrame:
    """
    Transforms specified columns of a dataframe into single columns where each row
    contains a dictionary of column names and values from the original dataframe.

    :param df: The dataframe to transform.
    :param groupings: A mapping that indicates which columns to collapse into dictionaries
        and what to call the new resulting column.
        If only a list of column names to be collapsed is given, it will be interpreted as
        the group_column_names in a single `{"collapsed": group_column_names}` dictionary,
        that is, all `group_column_names` are to be collapsed in to a single collapsed column.

    :return: A dataframe with the original columns not specified in `columns` untouched,
        and a new column `new_column_name` containing dictionaries of the collapsed columns.

    Example:

    >>> df = pd.DataFrame({
    ...     'a': [1, 1, 2, 2],
    ...     'b': [3, 4, 5, 6],
    ...     'c': [7, 8, 9, 10]
    ... })
    >>> df  # doctest: +NORMALIZE_WHITESPACE
       a  b   c
    0  1  3   7
    1  1  4   8
    2  2  5   9
    3  2  6  10
    >>> collapse_columns(df, {'ab': ['a', 'b']})  # doctest: +NORMALIZE_WHITESPACE
       c   ab
    0  7  {'a': 1, 'b': 3}
    1  8  {'a': 1, 'b': 4}
    2  9  {'a': 2, 'b': 5}
    3 10  {'a': 2, 'b': 6}
    """
    if isinstance(groupings, str):
        groupings = [groupings]
    if isinstance(groupings, list):
        # Convert list to a dictionary with default key if groupings is a list
        groupings = {'collapsed': groupings}

    # Copy the dataframe to avoid changing the original one
    result_df = df.copy()

    # Apply the grouping transformation
    for new_column, columns_to_collapse in groupings.items():
        # Ensure only valid columns are processed
        valid_columns = [col for col in columns_to_collapse if col in df.columns]

        if not valid_columns:
            raise ValueError(
                f'No valid columns found in {columns_to_collapse} to collapse into {new_column}.'
            )

        # Create a new column with dictionaries mapping old column names to their values
        result_df[new_column] = df[valid_columns].apply(
            lambda row: row.to_dict(), axis=1
        )

        # Remove the original columns that have been collapsed
        result_df.drop(columns=valid_columns, inplace=True)

    return result_df


def column_sep_key_mapper(key, column_name, sep: str):
    return f'{column_name}{sep}{key}'


_column_dot_key_mapper = partial(column_sep_key_mapper, sep='.')


def _apply_key_mapper_to_keys(d: Union[dict, list], column_name, key_mapper):
    if isinstance(d, dict):
        return {key_mapper(key, column_name): value for key, value in d.items()}
    else:
        return {key_mapper(i, column_name): value for i, value in enumerate(d)}


def expand_columns(
    df: pd.DataFrame,
    expand_columns: Union[str, List[str]],
    *,
    drop=True,
    key_mapper=_column_dot_key_mapper,
) -> pd.DataFrame:
    """
    Expands the iterable values of specified columns in to new columns.
    The new columns will be named using the column_name and the key of the values of
    the iterable that is expanded (key if dict, integer index if sequence).

    :param df: The dataframe to transform.
    :param expand_columns: A list of column names whose values are dictionaries
        that need to be expanded into new columns.
    :param drop: Whether to drop the original columns that were expanded.
    :param key_mapper: A function that takes a key and a column name and returns a
        new key. By default, the new key is the concatenation of the column name and
        the original key. If None, will just take the original key. The reason for
        also taking the column_name by default is to avoid collisions if the keys are
        used in more than one column.

    :return: A dataframe with the expanded columns added.

    Examples:
    >>> df = pd.DataFrame({
    ...     'c': [7, 8, 9, 10],
    ...     'X': [{'a': 1, 'b': 3}, {'a': 1, 'b': 4}, {'a': 2, 'b': 5}, {'a': 2, 'b': 6}]
    ... })
    >>> expand_columns(df, ['X'])  # doctest: +NORMALIZE_WHITESPACE
       c  X.a  X.b
    0  7  1  3
    1  8  1  4
    2  9  2  5
    3 10  2  6

    Let's see what happens when the elements of an expanded column are lists instead of
    dicts, we ask to not drop, and we use `key_mapper=None`:

    >>> df = pd.DataFrame({
    ...     'c': [7, 8, 9, 10],
    ...     'X': [[1, 3], [1, 4], [2, 5], [2, 6]]
    ... })
    >>> expand_columns(df, ['X'], drop=False, key_mapper=None)  # doctest: +NORMALIZE_WHITESPACE
        c       X  0  1
    0   7  [1, 3]  1  3
    1   8  [1, 4]  1  4
    2   9  [2, 5]  2  5
    3  10  [2, 6]  2  6

    """
    if isinstance(expand_columns, str):
        expand_columns = [expand_columns]
    # Validate that all expand_columns are in df.columns
    for col in expand_columns:
        if col not in df.columns:
            raise ValueError(f'Column {col} does not exist in the DataFrame.')
    if key_mapper is None:
        key_mapper = lambda key, column_name: key

    # Copy the dataframe to avoid changing the original one
    result_df = df.copy()

    # Apply the expansion transformation
    for col in expand_columns:
        _map_keys_of_dict = partial(
            _apply_key_mapper_to_keys, column_name=col, key_mapper=key_mapper
        )
        apply_func = lambda x: pd.Series(_map_keys_of_dict(x))
        # Extract dictionary to separate columns
        new_cols = result_df[col].apply(apply_func)
        # Merge new columns into the result DataFrame
        result_df = pd.concat([result_df, new_cols], axis=1)
        # Optionally drop the original column that was expanded
        if drop:
            result_df.drop(columns=[col], inplace=True)

    return result_df



"""
A  (key-value) data-object-layer to get (pandas) tables from a variety of sources with ease.

"""

from tabled.html import get_tables_from_url
from tabled.base import (
    DfFiles,
    DfLocalFileReader,  # will be deprecated
    KeyFuncReader,
    DataframeKvReader,
    dflt_ext_mapping,
    df_from_data_according_to_key,
    get_file_ext,
    identity,
    key_func_mapping,
    split_keys,
)
from tabled.multi import ColumnOrientedMapping
from tabled.util import (
    collapse_rows,
    expand_rows,
    collapse_columns,
    expand_columns,
    auto_decode_bytes,
)



import pandas as pd
import pytest
from join_tables import *


@pytest.fixture
def simple_tables():
    return {
        'A': pd.DataFrame({'x': [1, 2], 'y': [3, 4]}),
        'B': pd.DataFrame({'y': [3, 4], 'z': [5, 6]}),
    }


@pytest.fixture
def complex_tables():
    return {
        'A': pd.DataFrame({'a': [1, 2], 'b': [3, 4], 'c': [5, 6]}),
        'B': pd.DataFrame({'b': [3, 4], 'd': [7, 8]}),
        'C': pd.DataFrame({'d': [7, 8], 'e': [9, 10], 'f': [11, 12]}),
        'D': pd.DataFrame({'f': [11, 12], 'g': [13, 14]}),
    }


@pytest.fixture
def tables():
    return {
        'A': pd.DataFrame({'b': [1, 2, 3, 33], 'c': [4, 5, 6, 66]}),
        'B': pd.DataFrame(
            {
                'b': [1, 2, 3],
                'a': [4, 5, 6],
                'd': [7, 8, 9],
                'e': [10, 11, 12],
                'f': [13, 14, 15],
            }
        ),
        'C': pd.DataFrame({'f': [13, 14, 15], 'g': [4, 5, 6]}),
        'D': pd.DataFrame(
            {'d': [7, 8, 77], 'e': [10, 11, 77], 'h': [7, 8, 9], 'i': [1, 2, 3]}
        ),
        'E': pd.DataFrame({'i': [1, 2, 3], 'j': [4, 5, 6]}),
    }


# Test for the simple case
def test_minimum_covering_tree_simple(simple_tables):
    target_subset = ['y', 'z']
    edges = minimum_covering_tree(simple_tables, target_subset)
    assert set(edges) == {
        ('A', 'B')
    }, f"Expected edges: {set([('A', 'B')])}, but got: {set(edges)}"


def test_generate_join_sequence_simple(simple_tables):
    target_subset = {'y', 'z'}
    leaf_edges = get_leaf_edges(simple_tables, target_subset)
    join_sequence = generate_join_sequence(simple_tables, leaf_edges, target_subset)
    expected_sequence = ['A', Join('B', remove=['x'])]
    assert (
        join_sequence == expected_sequence
    ), f'Expected join sequence: {expected_sequence}, but got: {join_sequence}'


def test_compute_join_resolution_simple(simple_tables):
    target_subset = {'y', 'z'}
    leaf_edges = get_leaf_edges(simple_tables, target_subset)
    join_sequence = generate_join_sequence(simple_tables, leaf_edges, target_subset)
    expected_result = pd.DataFrame({'y': [3, 4], 'z': [5, 6]})
    result = compute_join_resolution(join_sequence, simple_tables)
    pd.testing.assert_frame_equal(result, expected_result)


# Test for the complex case
# def test_minimum_covering_tree_complex(complex_tables):
#     target_subset = ['a', 'e', 'g']
#     edges = minimum_covering_tree(complex_tables, target_subset)
#     expected_edges = {('A', 'B'), ('B', 'C'), ('C', 'D')}
#     assert set(edges) == expected_edges, f"Expected edges: {expected_edges}, but got: {set(edges)}"

# def test_generate_join_sequence_complex(complex_tables):
#     target_subset = {'a', 'e', 'g'}
#     leaf_edges = get_leaf_edges(complex_tables, target_subset)
#     join_sequence = generate_join_sequence(complex_tables, leaf_edges, target_subset)
#     expected_sequence = ['A', Join('B', remove=['c']), Join('C', remove=['b']), Join('D', remove=['d', 'f'])]
#     print(join_sequence)
#     assert join_sequence == expected_sequence, f"Expected join sequence: {expected_sequence}, but got: {join_sequence}"

# def test_compute_join_resolution_complex(complex_tables):
#     target_subset = {'a', 'e', 'g'}
#     leaf_edges = get_leaf_edges(complex_tables,target_subset)
#     join_sequence = generate_join_sequence(complex_tables, leaf_edges, target_subset)
#     expected_result = pd.DataFrame({'a': [1, 2], 'e': [9, 10], 'g': [13, 14]})
#     result = compute_join_resolution(join_sequence, complex_tables)
#     pd.testing.assert_frame_equal(result, expected_result)

# Test for tables
def test_minimum_covering_tree(tables):
    target_subset = {'b', 'g', 'j'}
    edges = minimum_covering_tree(tables, target_subset)
    expected_edges = {('A', 'B'), ('B', 'C'), ('B', 'D'), ('D', 'E')}
    assert (
        set(edges) == expected_edges
    ), f'Expected edges: {expected_edges}, but got: {set(edges)}'


def test_generate_join_sequence(tables):
    target_subset = {'b', 'g', 'j'}
    leaf_edges = get_leaf_edges(tables, target_subset)
    join_sequence = generate_join_sequence(tables, leaf_edges, target_subset)
    expected_sequence = [
        'B',
        Join('C', remove=['f', 'a']),
        Join('D', remove=['e', 'd', 'h']),
        Join('E', remove=['i']),
    ]
    assert (
        join_sequence == expected_sequence
    ), f'Expected join sequence: {expected_sequence}, but got: {join_sequence}'


def test_compute_join_resolution(tables):
    target_subset = {'b', 'g', 'j'}
    leaf_edges = get_leaf_edges(tables, target_subset)
    join_sequence = generate_join_sequence(tables, leaf_edges, target_subset)
    expected_result = pd.DataFrame({'b': [1, 2], 'g': [4, 5], 'j': [4, 5]})
    result = compute_join_resolution(join_sequence, tables)
    pd.testing.assert_frame_equal(result, expected_result)


from tabled.multi import execute_commands
from tabled.multi import execute_table_commands


def test_execute_commands_simply():
    from tabled.multi import Join, Remove, Load, Rename

    # ---------------------------------------------
    # First silly test

    silly_interpreter_map = {
        Join: lambda scope, command: f'Joining {command.table_key}',
        Remove: lambda scope, command: f'Removing {command.fields}',
    }

    g = execute_commands(
        [Join('asdf'), Remove('apple')], scope={}, interpreter_map=silly_interpreter_map
    )
    assert list(g) == ['Joining asdf', 'Removing apple']

    # ---------------------------------------------
    # The real case

    table1 = pd.DataFrame({'ID': [1, 2, 3], 'Name': ['Alice', 'Bob', 'Charlie']})
    table2 = pd.DataFrame({'ID': [2, 3, 4], 'Age': [25, 30, 22]})
    table3 = pd.DataFrame({'ID': [1, 2, 3, 4], 'Salary': [50000, 60000, 70000, 55000]})

    tables = {'table1': table1, 'table2': table2, 'table3': table3}

    commands = [
        Load('table1'),
        Rename({'Name': 'First Name'}),
        Remove(['First Name']),
        Join('table3'),
    ]

    scope = tables
    extra_scope = dict()

    it = execute_table_commands(commands, tables, extra_scope=extra_scope)

    def are_equal(a, b):
        if isinstance(a, pd.DataFrame) and isinstance(b, pd.DataFrame):
            return (a == b).all().all()
        else:
            return a == b

    next(it)
    assert are_equal(extra_scope['cumul'], scope['table1'])
    next(it)
    assert list(extra_scope['cumul']) == ['ID', 'First Name']
    next(it)
    assert are_equal(extra_scope['cumul'], pd.DataFrame({'ID': [1, 2, 3]}))
    next(it)
    assert list(extra_scope['cumul']) == ['ID', 'Salary']


# Test wiki table
from tabled.html import *


def test_extract_wikipedia_tables():
    wikiurl = 'https://fr.wikipedia.org/wiki/Liste_des_communes_de_France_les_plus_peupl%C3%A9es'
    converter = url_to_html_func('requests')
    tables = get_tables_from_url(wikiurl, url_to_html=converter)
    assert tables is not None
    assert len(tables) > 0
    for idx, df in enumerate(tables):
        assert isinstance(df, pd.DataFrame)
        assert not df.empty


@pytest.fixture
def extracted_dataframes():
    converter = url_to_html_func('requests')
    url_aeroports_frequentes = 'https://fr.wikipedia.org/wiki/Liste_des_a%C3%A9roports_les_plus_fr%C3%A9quent%C3%A9s_en_France'
    url_aeroports_vastes = 'https://fr.wikipedia.org/wiki/Liste_des_a%C3%A9roports_les_plus_vastes_au_monde'

    dfs_aeroports_frequentes = get_tables_from_url(
        url_aeroports_frequentes, url_to_html=converter
    )
    dfs_aeroports_vastes = get_tables_from_url(
        url_aeroports_vastes, url_to_html=converter
    )

    return dfs_aeroports_frequentes, dfs_aeroports_vastes


def test_execute_commands_wiki(extracted_dataframes):
    from tabled.multi import Join, Remove, Load, Rename

    table1_wiki = extracted_dataframes[0][0]
    table2_wiki = extracted_dataframes[1][0]

    tables = {'table1_wiki': table1_wiki, 'table2_wiki': table2_wiki}
    commands = [
        Load('table2_wiki'),
        Remove('Aéroport'),
        Rename({'Code': 'Code IATA'}),
        Join('table1_wiki'),
    ]

    scope = tables
    extra_scope = dict()
    it = execute_table_commands(commands, scope, extra_scope=extra_scope)
    next(it)
    next(it)
    next(it)
    next(it)
    assert extra_scope['cumul'].shape[0] == 1



"""A module containing datas that will be used for tests"""

ID = 'idx'

# Simple with one field having a repeated value
three_simple_docs = [
    {ID: 0, 's': 'a', 'n': 1},
    {ID: 1, 's': 'b', 'n': 2},
    {ID: 2, 's': 'b', 'n': 3},
]

# Schema not "perfect": Some missing fields here and there.
# Good to test no-sql cases
nums_and_lans = [
    {ID: 1, 'en': 'one', 'fr': 'un', 'sp': 'uno', 'so_far': [1]},
    {ID: 2, 'en': 'two', 'fr': 'deux', 'so_far': [1, 2]},  # sp missing
    {ID: 3, 'en': 'three', 'fr': 'trois', 'sp': 'tres', 'so_far': [1, 2, 3]},
    {ID: 4, 'en': 'four', 'fr': 'quatre', 'sp': 'cuatro', 'so_far': [1, 2, 3, 4],},
    {ID: 5, 'en': 'five', 'sp': 'cinco', 'so_far': [1, 2, 3, 4, 5],},  # fr missing
]

# Stable schema
# Groupby possibilities (see number: Several unique values) -- this allows to test filtering more naturally
feature_cube = [
    {ID: 1, 'number': 6, 'color': 'red', 'dims': {'x': 2, 'y': 3}},
    {ID: 2, 'number': 6, 'color': 'blue', 'dims': {'x': 3, 'y': 2}},
    {ID: 3, 'number': 10, 'color': 'red', 'dims': {'x': 2, 'y': 5}},
    {ID: 4, 'number': 10, 'color': 'red', 'dims': {'x': 5, 'y': 2}},
    {ID: 5, 'number': 15, 'color': 'red', 'dims': {'x': 3, 'y': 5}},
    {ID: 6, 'number': 15, 'color': 'blue', 'dims': {'x': 3, 'y': 5}},
    {ID: 7, 'number': 15, 'color': 'blue', 'dims': {'x': 5, 'y': 3}},
]

# Sequence Annotations
# Demoing stream annotations with source and time interval
sequence_annots = [
    {'source': 'audio', 'bt': 5, 'tt': 7, 'annot': 'cat'},
    {
        'source': 'audio',
        'bt': 6,
        'tt': 9,
        'annot': 'dog',
        'comments': 'barks and chases cat away',
    },
    {'source': 'visual', 'bt': 5, 'tt': 8, 'annot': 'cat'},
    {
        'source': 'visual',
        'bt': 6,
        'tt': 15,
        'annot': 'dog',
        'comments': 'dog remains in view after bark ceases',
    },
]

####### Rock bands #####################################################################################################

rock_bands = [
    {
        ID: 'pink_floyd_id',
        'name': 'Pink Floyd',
        'members': [
            {'_id': '1', 'firstname': 'Roger', 'lastname': 'Waters'},
            {'_id': '2', 'firstname': 'Nick', 'lastname': 'Mason'},
            {'_id': '3', 'firstname': 'Syd', 'lastname': 'Barrett'},
            {'_id': '4', 'firstname': 'Richard', 'lastname': 'Write'},
        ],
    },
    {
        ID: 'the_doors_id',
        'name': 'The Doors',
        'members': [
            {'_id': '1', 'firstname': 'Jim', 'lastname': 'Morrison'},
            {'_id': '2', 'firstname': 'Ray', 'lastname': 'Manzarek'},
            {'_id': '3', 'firstname': 'Robby', 'lastname': 'Krieger'},
            {'_id': '4', 'firstname': 'John', 'lastname': 'Densmore'},
        ],
    },
]

####### BDFLs ##########################################################################################################
# ubtained with:
# from scrapyng.tables import get_tables_from_url (if no scrapyng, use ut.webscrape.tables)
# t = get_tables_from_url('https://en.wikipedia.org/wiki/Benevolent_dictator_for_life')[0]
# t = t[['Name', 'Project', 'Type']].to_dict(orient='records')

bdfl = [
    {
        'Name': 'Sylvain Benner',
        'Project': 'Spacemacs',
        'Type': 'Community-driven Emacs distribution',
    },
    {
        'Name': 'Vitalik Buterin',
        'Project': 'Ethereum',
        'Type': 'Blockchain-based cryptocurrency',
    },
    {
        'Name': 'Dries Buytaert',
        'Project': 'Drupal',
        'Type': 'Content management framework',
    },
    {'Name': 'Haoyuan Li', 'Project': 'Alluxio', 'Type': 'Data Orchestration System',},
    {
        'Name': 'Evan Czaplicki',
        'Project': 'Elm',
        'Type': 'Front-end web programming language',
    },
    {
        'Name': 'David Heinemeier Hansson',
        'Project': 'Ruby on Rails',
        'Type': 'Web framework',
    },
    {'Name': 'Rich Hickey', 'Project': 'Clojure', 'Type': 'Programming language',},
    {
        'Name': 'Adrian Holovaty and Jacob Kaplan-Moss',
        'Project': 'Django',
        'Type': 'Web framework',
    },
    {
        'Name': 'Laurent Destailleur',
        'Project': 'Dolibarr ERP CRM',
        'Type': 'Software suite for Enterprise Resource Planning and Customer '
        'Relationship Management',
    },
    {
        'Name': 'Francois Chollet',
        'Project': 'Keras',
        'Type': 'Deep learning framework',
    },
    {'Name': 'Xavier Leroy', 'Project': 'OCaml', 'Type': 'Programming language',},
    {
        'Name': 'Yukihiro Matsumoto (Matz)',
        'Project': 'Ruby',
        'Type': 'Programming language',
    },
    {
        'Name': 'Wes McKinney',
        'Project': 'Pandas',
        'Type': 'Python data analysis library',
    },
    {'Name': 'Bram Moolenaar', 'Project': 'Vim', 'Type': 'Text editor'},
    {
        'Name': 'Matt Mullenweg [a]',
        'Project': 'WordPress',
        'Type': 'Content management framework',
    },
    {'Name': 'Martin Odersky', 'Project': 'Scala', 'Type': 'Programming language',},
    {'Name': 'Taylor Otwell', 'Project': 'Laravel', 'Type': 'Web framework'},
    {
        'Name': 'Theo de Raadt',
        'Project': 'OpenBSD',
        'Type': 'A Unix-like operating system',
    },
    {
        'Name': 'Ton Roosendaal[b]',
        'Project': 'Blender',
        'Type': '3D computer graphics software',
    },
    {
        'Name': 'Sébastien Ros',
        'Project': 'Orchard Project',
        'Type': 'Content management system',
    },
    {
        'Name': 'Mark Shuttleworth[c]',
        'Project': 'Ubuntu',
        'Type': 'Linux distribution',
    },
    {'Name': 'Don Syme[d]', 'Project': 'F#', 'Type': 'Programming language'},
    {
        'Name': 'Linus Torvalds[e]',
        'Project': 'Linux',
        'Type': 'Operating system kernel',
    },
    {'Name': 'José Valim', 'Project': 'Elixir', 'Type': 'Programming language',},
    {
        'Name': 'Pauli Virtanen',
        'Project': 'SciPy',
        'Type': 'Python library used for scientific and technical computing',
    },
    {
        'Name': 'Patrick Volkerding',
        'Project': 'Slackware',
        'Type': 'GNU/Linux distribution',
    },
    {
        'Name': 'Nathan Voxland',
        'Project': 'Liquibase',
        'Type': 'Database schema management',
    },
    {
        'Name': 'Shaun Walker',
        'Project': 'DotNetNuke',
        'Type': 'Web application framework',
    },
    {'Name': 'Larry Wall', 'Project': 'Perl', 'Type': 'Programming language'},
    {'Name': 'Jeremy Soller[37]', 'Project': 'Redox', 'Type': 'Operating system',},
    {
        'Name': 'Eugen Rochko',
        'Project': 'Mastodon',
        'Type': 'open source, decentralized social network',
    },
    {
        'Name': 'Dylan Araps',
        'Project': 'KISS Linux',
        'Type': 'a bare-bones Linux distribution based on musl libc and BusyBox',
    },
    {
        'Name': 'Gavin Mendel-Gleason[f]',
        'Project': 'TerminusDB',
        'Type': 'Open-source graph database for knowledge graph representation',
    },
]



"""Multi-tabled data structures."""

from functools import cached_property
import pandas as pd
from typing import TypeVar, Callable, KT, Union, Iterable, Mapping, Dict

Column = TypeVar('Column')
TableKey = TypeVar('TableKey')
MappingOfDataFrames = Mapping[KT, pd.DataFrame]
DataFrames = Union[MappingOfDataFrames, Iterable[pd.DataFrame]]

# --------------------------------------------------------------------------------------
# Utils


def mapping_of_dataframes(tables: DataFrames) -> MappingOfDataFrames:
    """Cast to a mapping of dataframes"""
    if not isinstance(tables, Mapping):
        if isinstance(tables, Iterable):
            tables = dict(enumerate(tables))
        else:
            raise TypeError(f'Expected Mapping or Iterable, got {type(tables)}')
    return tables


def dataframes(tables: DataFrames) -> Iterable[pd.DataFrame]:
    """Cast to an iterable of dataframes."""
    if isinstance(tables, Mapping):
        tables = tables.values()
    return tables


# --------------------------------------------------------------------------------------
# Combined datasets
# See https://github.com/i2mint/tabled/discussions/3
from dataclasses import dataclass
from collections import ChainMap


def execute_commands(
    commands: Iterable, scope: Mapping, interpreter_map: Mapping, *, extra_scope=None,
):
    """
    Carries `commands` operations out with tables taken from `scope`.

    :param commands: An iterable of join operations to carry out.
    """
    if extra_scope is None:
        extra_scope = dict()
    _scope = ChainMap(extra_scope, scope)

    for command in commands:
        command_type = type(command)
        command_executor = interpreter_map.get(command_type, None)
        if command_executor is None:
            raise TypeError(f'Unknown command type: {type(command)}')
        yield command_executor(_scope, command)


# Define the JoinWith dataclass
# @dataclass
# class JoinWith:
#     table_key: str
#     remove: list = None


@dataclass
class Load:
    key: str


@dataclass
class Join:
    table_key: str


@dataclass
class Remove:
    fields: Union[str, Iterable[str]]


@dataclass
class Rename:
    rename_mapping: Dict[str, str]


def set_scope_value(scope, key, value):
    scope[key] = value


def load_func(scope, command):
    return set_scope_value(scope, 'cumul', scope[command.key])


def join_func(scope, command):
    table = scope[command.table_key].copy()
    if 'renamed_columns' in scope:
        for old_col, new_col in scope['renamed_columns'].items():
            if old_col in table:
                table.rename(columns={old_col: new_col}, inplace=True)
    cumul = scope['cumul']
    scope['cumul'] = cumul.merge(table, how='inner')


def remove_func(scope, command):
    scope['cumul'] = scope['cumul'].drop(columns=command.fields)


def rename_func(scope, command):
    scope['renamed_columns'] = command.rename_mapping
    for old_col, new_col in command.rename_mapping.items():
        scope['cumul'] = scope['cumul'].rename(columns={old_col: new_col})


dflt_tables_interpreter_map = {
    Load: load_func,
    Join: join_func,
    Remove: remove_func,
    Rename: rename_func,
}


def execute_table_commands(
    commands: Iterable,
    tables: Mapping[str, pd.DataFrame],
    interpreter_map: Mapping = dflt_tables_interpreter_map,
    *,
    extra_scope=None,
):
    return execute_commands(commands, tables, interpreter_map, extra_scope=extra_scope)


# def execute_commands(
#     commands: Iterable, tables: Mapping[str, pd.DataFrame]
# ) -> pd.DataFrame:
#     """
#     Carries `commands` operations out with tables taken from `tables`.

#     :param commands: An iterable of join operations to carry out.
#         Each join operation is either a table name (str) or a JoinWith object.
#         If it's a JoinWith object, it's assumed that the table has already been joined
#         and the fields to remove are in the `remove` attribute of the object.
#     :param tables: A mapping of table names to tables (pd.DataFrame)
#     """
#     # join_ops = map(ensure_join_op, resolution_sequence)
#     commands = iter(commands)
#     first_command = next(commands)
#     assert isinstance(first_command, Join)
#     table_key = first_command.table_key
#     cumul = tables[table_key]  # initialize my accumulator
#     for command in commands:
#         if isinstance(command, Join):
#             table = tables[command.table_key]
#             cumul = cumul.merge(table, how='inner')
#         elif isinstance(command, Remove):
#             cumul = cumul.drop(columns=command.fields)
#         else:
#             raise TypeError(f'Unknown command type: {type(command)}')
#     return cumul


# --------------------------------------------------------------------------------------
# Row-oriented mapping of multiple tables


# --------------------------------------------------------------------------------------
# Column-oriented mapping of multiple tables


def columns_of_first_table(tables: MappingOfDataFrames) -> Iterable[Column]:
    tables = mapping_of_dataframes(tables)
    first_table = next(iter(dataframes(tables)))
    return first_table.columns.values.tolist()


def columns_of_all_tables(tables: MappingOfDataFrames) -> Iterable[Column]:
    """Return all columns from all tables, in order of first appearance.
    This is useful for making a ColumnOrientedMapping without reverting to
    the default columns argument, which is to use the columns of the first table.
    """
    from itertools import chain

    tables = mapping_of_dataframes(tables)
    all_columns = chain.from_iterable(table.columns for table in tables.values())
    # return unique columns in order of first appearance
    return list(dict.fromkeys(all_columns))


class ColumnOrientedMapping(Mapping):
    def __init__(
        self,
        tables: MappingOfDataFrames,
        columns: Union[Callable, Iterable[Column]] = columns_of_first_table,
    ):
        self.tables = mapping_of_dataframes(tables)
        self._init_columns = columns

    @cached_property
    def columns(self):
        """The columns that will be used in this mapping (the keys of the mapping)"""
        if isinstance(self._init_columns, str):
            return [self._init_columns]
        elif callable(self._init_columns):
            return self._init_columns(self.tables)
        assert isinstance(
            self._init_columns, Iterable
        ), f'Expected Callable or Iterable, got {self._init_columns}'
        return self._init_columns

    @cached_property
    def _table_keys(self) -> Iterable[TableKey]:
        return list(self.tables.keys())

    def __iter__(self) -> Iterable[Column]:
        return iter(self.columns)

    def __getitem__(self, k):
        """Return a table with the given columns from all tables."""
        return self.df(k)

    def __len__(self):
        return len(self.columns)

    def __contains__(self, k):
        return k in self.columns

    def df(self, columns=None):
        """Concatinate all dataframes of given columns from all tables,
        returning a single dataframe."""
        if columns is None:
            columns = self.columns
        return pd.concat([table[columns] for table in dataframes(self.tables)])

    def array(self, columns=None):
        """Concatinate all the arrays of given columns from all tables,
        returning a single array."""
        return self.df(columns).array


# Convenience functions, placed in ColumnOrientedMapping for easy access
ColumnOrientedMapping.columns_of_first_table = columns_of_first_table
ColumnOrientedMapping.columns_of_all_tables = columns_of_all_tables



"""Based functionality for tabled"""

# TODO: Completely change so that
#   * routing is more general (not only conditioned on ext, but function of key,
#   and possibly data
#   * Allow on-the-fly parametrization (e.g. sep='\t', index_col=False...)
from functools import partial
from io import BytesIO
from typing import (
    Mapping,
    Callable,
    TypeVar,
    KT,
    VT,
    TypeVar,
    Dict,
    Tuple,
    Iterable,
    Union,
)
import os
import re
import pickle

import pandas as pd
from i2 import mk_sentinel, Sig

from dol import Files  # previously: py2store.stores.local_store import LocalBinaryStore
from tabled.util import identity, split_keys


Obj = TypeVar('Obj')
KeyFunc = Callable[[Obj], KT]
Extension = TypeVar('Extension')
DfDecoder = Callable[[Obj], pd.DataFrame]
dflt_not_found_sentinel = mk_sentinel('dflt_not_found_sentinel')


def key_func_mapping(
    obj: Obj,
    mapping: Mapping[KT, VT],
    key: KeyFunc = identity,
    not_found_sentinel=dflt_not_found_sentinel,
) -> VT:
    """Map an object to a value based on a key function"""
    return mapping.get(key(obj), not_found_sentinel)


from dol import KvReader


class KeyFuncReader(KvReader):
    def __init__(self, mapping: Mapping[KT, VT], key: KeyFunc = identity):
        self.mapping = mapping
        self.key = key

    def __getitem__(self, k):
        return self.mapping[self.key(k)]

    def __contains__(self, k):
        return self.key(k) in self.mapping

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

    def __repr__(self):
        return f'{type(self).__name__}({self.mapping}, key={self.key})'


# TODO: Add some registry functionality to this?
# TODO: Merge with dol's extension-based codec mapping (routing)
dflt_ext_mapping = split_keys(
    {
        'xls xlsx xlsm': partial(pd.read_excel, index=False),  # Excel files
        'csv txt': partial(pd.read_csv, index_col=False),  # CSV and text files
        'tsv': partial(pd.read_csv, sep='\t', index_col=False),  # Tab-separated
        'parquet': pd.read_parquet,  # Parquet format
        'json': partial(pd.read_json, orient='records'),  # JSON format
        'html': partial(pd.read_html, index_col=False),  # HTML tables
        'p pickle pkl': pickle.load,  # Pickle files
        'xml': pd.read_xml,  # XML files
        'h5 hdf5': pd.read_hdf,  # HDF5 format
        'sql sqlite': pd.read_sql,  # SQL queries
        'feather': pd.read_feather,  # Feather format
        'stata dta': pd.read_stata,  # Stata files
        'sas': pd.read_sas,  # SAS files
        # 'gbq': pandas_gbq.read_gbq,  # Google BigQuery
    }
)


def df_from_data_given_ext(
    data, ext: Extension, mapping=dflt_ext_mapping, **extra_decoder_kwargs
):
    """Get a dataframe from a (data, ext) pair"""
    if ext.startswith('.'):
        ext = ext[1:]
    decoder = key_func_mapping(
        ext, mapping, key=identity, not_found_sentinel=None,  # TODO
    )
    if decoder is not None:
        # pluck out any key-value pairs of extra_decoder_kwargs whose names are
        # arguments of decoder. (This is needed since extra_decoder_kwargs can be
        # servicing multiple decoders, and we don't want to pass arguments to the
        # wrong decoder.)
        extra_decoder_kwargs = Sig(decoder).map_arguments(
            (), extra_decoder_kwargs, allow_partial=True, allow_excess=True
        )
        # decode the data
        return decoder(data, **extra_decoder_kwargs)
    else:
        raise ValueError(f"Don't know how to handle extension: {ext}")


def df_from_data_according_to_key(
    data: Obj, mapping: Dict[Extension, DfDecoder], key: KT, **extra_decoder_kwargs
):
    """Get a dataframe from a (data, mapping, key) triple"""
    decoder = key_func_mapping(data, mapping, key=key, not_found_sentinel=None)
    return decoder(data, **extra_decoder_kwargs)


def get_file_ext(key: KT) -> Extension:
    _, ext = os.path.splitext(key)
    if ext:
        return ext[1:].lower()
    else:
        return ext


protocol_re = re.compile(r'([a-zA-Z0-9]+)://')


def get_protocol(url: str):
    """Get the protocol of a url

    >>> get_protocol('https://www.google.com')
    'https'
    >>> get_protocol('file:///home/user/file.txt')
    'file'

    The function returns None if no protocol is found:

    >>> assert get_protocol('no_protocol_here') is None

    """
    m = protocol_re.match(url)
    if m:
        return m.group(1)


df_from_data_according_to_ext = partial(
    df_from_data_according_to_key, mapping=dflt_ext_mapping, key=get_file_ext,
)

# df_from_data_given_ext meant to be equivalent (but more general, using ext_specs) to
# def df_from_data_given_ext(data, ext, ext_specs=None, **kwargs):
#     """Get a dataframe from a (data, ext) pair"""
#     ext_specs = ext_specs or DFLT_EXT_SPECS  # NOTE: Note used yet
#     if ext.startswith("."):
#         ext = ext[1:]
#     if ext in {"xls", "xlsx"}:
#         kwargs = dict({"index": False}, **kwargs)
#         return pd.read_excel(data, **kwargs)
#     elif ext in {"csv"}:
#         kwargs = dict({"index_col": False}, **kwargs)
#         return pd.read_csv(data, **kwargs)
#     elif ext in {"tsv"}:
#         kwargs = dict({"sep": "\t", "index_col": False}, **kwargs)
#         return pd.read_csv(data, **kwargs)
#     elif ext in {"json"}:
#         kwargs = dict({"orient": "records"}, **kwargs)
#         return pd.read_json(data, **kwargs)
#     elif ext in {"html"}:
#         kwargs = dict({"index_col": False}, **kwargs)
#         return pd.read_html(data, **kwargs)[0]
#     elif ext in {"p", "pickle"}:
#         return pickle.load(data, **kwargs)
#     else:
#         raise ValueError(f"Don't know how to handle extension: {ext}")


# TODO: Make the logic independent from local files assumption.
# TODO: Better separate Reader, and add DfStore to make a writer.
# TODO: Add filtering functionality in init? (By function, regex, extension?)
class DfFiles(Files):
    """A key-value store providing values as pandas.DataFrames.

    Use Case: You have a bunch of files in a folder, all corresponding to some
    dataframes that were saved in some way. You want to a key-value store whose values
    are the (decoded) dataframes corresponding to the files in the folder.

    Args:
    rootdir: A root directory.
    extension_decoder_mapping: A mapping from file extensions to functions that can
        read the dataframes
    extra_decoder_kwargs: Extra arguments to pass to the decoder functions.



    """

    def __init__(
        self,
        rootdir: str,
        extension_decoder_mapping: Dict[Extension, DfDecoder] = dflt_ext_mapping,
        extra_decoder_kwargs: Union[dict, Iterable] = (),
    ):
        super().__init__(rootdir)

        self.key_to_ext = get_file_ext
        extra_decoder_kwargs = dict(extra_decoder_kwargs)
        self.data_and_ext_to_df = partial(
            df_from_data_given_ext,
            mapping=extension_decoder_mapping,
            **extra_decoder_kwargs,
        )

    def __getitem__(self, k):
        ext = self.key_to_ext(k)
        data = BytesIO(super().__getitem__(k))
        return self.data_and_ext_to_df(data, ext)

    def __setitem__(self, k, v):
        raise NotImplementedError('This is a reader: No write operation allowed')

    def __delitem__(self, k):
        raise NotImplementedError('This is a reader: No delete operation allowed')


DfReader = DfFiles  # alias for back-compatibility: TODO: Issue warning on use
DfLocalFileReader = DfFiles  # back-compatibility: TODO: Issue warning on use


# ------------------------------------------------------------------------------
# Wrapping a DataFrame to provide a key-value interface

import pandas as pd
from collections.abc import Mapping


def validate_fields(df, key_fields, value_columns):
    all_columns = df.columns.tolist()
    all_index_levels = list(df.index.names)

    for field in key_fields:
        if field not in all_columns and field not in all_index_levels:
            raise ValueError(
                f'Field {field} not found in the DataFrame columns or index levels.'
            )

    for col in value_columns:
        if col not in all_columns:
            raise ValueError(f'Column {col} not found in the DataFrame.')


# TODO: Does it accommodate the case where I want to define my key_fields to be the
# UNNAMED index of the dataframe, plus a column.
# For example, consider (and add to the test) the case where I have df.set_index('A'),
# but remove the name "A" from the index. Now I can't say key_fields=['A', 'B'] anymore.
class DataframeKvReader(Mapping):
    """
    A class to wrap a DataFrame and provide a Mapping interface where keys are
    combinations of specified columns or index levels, and values are sub-dataframes of specified columns.

    Example usage:
    >>> df = pd.DataFrame({
    ...     'A': [1, 2, 1],
    ...     'B': [4, 5, 4],
    ...     'C': [7, 8, 9],
    ...     'D': [10, 11, 12]
    ... })
    >>> df
       A  B  C   D
    0  1  4  7  10
    1  2  5  8  11
    2  1  4  9  12
    >>> kv_reader = DataframeKvReader(df, ['A', 'B'], ['C', 'D'])
    >>> key = (1, 4)
    >>> kv_reader[key].reset_index(drop=True)  # doctest: +NORMALIZE_WHITESPACE
       C   D
    0  7  10
    1  9  12
    >>> list(kv_reader) == [(1, 4), (2, 5)]
    True

    But what if one (or more) of the key fields is an index level?
    The DataframeKvReader can handle that too:

    >>> df = df.set_index(['A'])
    >>> df  # doctest: +NORMALIZE_WHITESPACE
       B  C   D
    A
    1  4  7  10
    2  5  8  11
    1  4  9  12
    >>> kv_reader = DataframeKvReader(df, ['A', 'B'], ['C', 'D'])
    >>> key = (1, 4)
    >>> kv_reader[key].reset_index(drop=True)  # doctest: +NORMALIZE_WHITESPACE
       C   D
    0  7  10
    1  9  12
    >>> list(kv_reader) == [(1, 4), (2, 5)]
    True

    """

    def __init__(self, df, key_fields, value_columns=None):
        """
        Initialize the DataframeKvReader.

        Parameters:
        df (pd.DataFrame): The DataFrame to wrap.
        key_fields (str or list of str): Fields (columns or index levels) to use as keys.
        value_columns (str or list of str): Column(s) to use as values. Default is all columns.

        """
        if value_columns is None:
            value_columns = df.columns.tolist()
        validate_fields(df, key_fields, value_columns)
        self.key_fields = key_fields if isinstance(key_fields, list) else [key_fields]
        self.value_columns = (
            value_columns if isinstance(value_columns, list) else [value_columns]
        )

        # Check if key fields are all index levels
        self.is_index_key = all(field in df.index.names for field in self.key_fields)

        if not self.is_index_key:
            # If key fields are not all index levels, reset index and set new index
            self.df = (
                df.reset_index().set_index(self.key_fields, drop=False).sort_index()
            )
        else:
            # If key fields are all index levels, ensure the DataFrame is sorted by those key fields
            self.df = df.sort_index()

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            key = (key,)
        try:
            sub_df = self.df.loc[key, self.value_columns]
            if isinstance(sub_df, pd.Series):
                sub_df = sub_df.to_frame().T
            return sub_df
        except KeyError:
            raise KeyError(f'Key {key} not found')

    def __iter__(self):
        keys = self.df.index.drop_duplicates().tolist()
        return iter(keys)

    def __len__(self):
        return self.df.index.nunique()

    def __contains__(self, key):
        return key in self.df.index

    def __repr__(self):
        return f'{type(self).__name__}(df=<{len(self.df)} rows>, key_fields={self.key_fields}, value_columns={self.value_columns})'



import pandas as pd
from tabled.base import DataframeKvReader


# Test cases
def test_DataframeKvReader():

    df = pd.DataFrame(
        {'A': [1, 2, 1], 'B': [4, 5, 4], 'C': [7, 8, 9], 'D': [10, 11, 12]}
    )

    kv_readers = [
        DataframeKvReader(df, ['A', 'B'], ['C', 'D']),
        DataframeKvReader(df.set_index(['A', 'B']), ['A', 'B'], ['C', 'D']),
        DataframeKvReader(df.set_index('A'), ['A', 'B'], ['C', 'D']),
    ]

    for kv_reader in kv_readers:
        # Accessing values
        key = (1, 4)
        expected_df = pd.DataFrame([{'C': 7, 'D': 10}, {'C': 9, 'D': 12}], index=[0, 2])
        assert (
            kv_reader[key]
            .reset_index(drop=True)
            .equals(expected_df.reset_index(drop=True))
        )
        assert list(kv_reader) == [(1, 4), (2, 5)]



"""To work with html"""

from typing import Callable, Union, Mapping
import re
import pandas as pd


def url_to_html_func(kind='requests') -> Callable:
    """Get a url_to_html function of a given kind."""
    url_to_html = None
    if kind == 'requests':
        import requests  # pip install requests

        def url_to_html(url):
            r = requests.get(url)
            if r.status_code != 200:
                print(
                    f'An error occured. Returning the response object for you to analyze: {r}'
                )
                return r
            return r.content

    elif kind == 'chrome' or kind == 'selenium':

        from selenium import webdriver  # pip install selenium
        from time import sleep

        def url_to_html(url, wait=2):
            b = webdriver.Chrome()
            b.get(url)
            if isinstance(wait, (int, float)):
                sleep(wait)
            html = b.page_source
            b.close()
            return html

    else:
        raise ValueError(f'Unknown url_to_html value: {url_to_html}')
    assert callable(url_to_html), "Couldn't make a url_to_html function"

    return url_to_html


get_tables_from_html = pd.read_html


def get_tables_from_url(
    url, *, url_to_html: Union[Callable, str] = 'requests', **tables_from_html_kwargs
):
    """Get's a list of pandas dataframes from tables scraped from a url.
    Note that this will only work with static pages. If the html needs to be rendered dynamically,
    you'll have to get your needed html otherwise (like with selenium).

    >>> url = 'https://en.wikipedia.org/wiki/List_of_musical_instruments'
    >>> tables = get_tables_from_url(url)  # doctest: +SKIP

    If you install selenium and download a chromedriver,
    you can even use your browser to render dynamic html.
    Say, to get updated coronavirus stats without a need to figure out the API
    (I mean, why have to figure out the language of an API, when someone already did that
    for you in their webpage!!):

    ```python
    url = 'https://www.worldometers.info/coronavirus/?utm_campaign=homeAdvegas1?'
    tables = get_tables_from_url(url, url_to_html='chrome')  # doctest: +SKIP
    ```

    To make selenium work:
    ```
        pip install selenium
        Download seleniumdriver here: https://chromedriver.chromium.org/
        Uzip and put in a place that's on you PATH (run command `echo $PATH` for a list of those places)
    ```
    """
    if not callable(url_to_html):
        url_to_html = url_to_html_func(url_to_html)
    try:
        return get_tables_from_html(url_to_html(url), **tables_from_html_kwargs)
    except ValueError as e:
        if len(e.args) > 0:
            msg, *_ = e.args
            if 'No tables found' in msg:
                return []
        raise


HTML_TEMPLATE1 = '''
<html>
<head>
<style>
  h2 {
    text-align: center;
    font-family: Helvetica, Arial, sans-serif;
  }
  table { 
    margin-left: auto;
    margin-right: auto;
  }
  table, th, td {
    border: 1px solid black;
    border-collapse: collapse;
  }
  th, td {
    padding: 5px;
    text-align: center;
    font-family: Helvetica, Arial, sans-serif;
    font-size: 90%;
  }
  table tbody tr:hover {
    background-color: #dddddd;
  }
  .wide {
    width: 90%; 
  }
</style>
</head>
<body>
'''

HTML_TEMPLATE2 = '''
</body>
</html>
'''


def df_to_html(df, title=None):
    ht = ''
    if title is not None:
        ht += f'<h2> {title} </h2>\n'
    ht += df.to_html(classes='wide', escape=False)
    return ht


def df_store_to_html(df_store, sep='\n<br>\n'):
    ht = ''
    for k, df in df_store.items():
        title = re.match(r'[^\d]+', k).group(0)
        ht += df_to_html(df, title)
        ht += sep
    return ht


def dfs_to_html_pretty(dfs, title=None):
    """
    Write an entire dataframe to an HTML file
    with nice formatting.
    Thanks to @stackoverflowuser2010 for the
    pretty printer see https://stackoverflow.com/a/47723330/362951
    """

    if isinstance(dfs, pd.DataFrame):
        ht = df_to_html(dfs, title=title)
    elif isinstance(dfs, Mapping):
        ht = df_store_to_html(dfs)
    else:
        ht = df_store_to_html(dict(enumerate(dfs)))

    return HTML_TEMPLATE1 + ht + HTML_TEMPLATE2


def dfs_to_pdf_bytes(dfs, title=None):
    import weasyprint

    html = dfs_to_html_pretty(dfs, title)
    return weasyprint.HTML(string=html).write_pdf()



from importlib.resources import files

project_name = 'tabled'
root_posix_path = files(project_name)


def local_posix(*path):
    return root_posix_path.joinpath(*path)


def local_text(*path):
    return local_posix(*path).read_text()


test_data_path = ('tests', 'data')
test_data_posix = local_posix(*test_data_path)


def test_data(name, text=False):
    name_posix = test_data_posix.joinpath(name)
    if not text:
        return name_posix.read_bytes()
    else:
        return name_posix.read_text()
